# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import os
from collections.abc import Iterator
from pathlib import Path

from pylint.testutils.functional.test_file import FunctionalTestFile

REASONABLY_DISPLAYABLE_VERTICALLY = 40
"""'Wet finger' number of python files that are reasonable to have in a functional test
directory.

'Wet finger' as in 'in my settings there are precisely this many'.
Initially the total number of files then we started counting only the python files to
avoid moving a lot of files.
"""

IGNORED_PARENT_DIRS = {
    "deprecated_relative_import",
    "ext",
    "regression",
    "regression_02",
    "regression_03",
    "used_02",
}
"""Direct parent directories that should be ignored."""

IGNORED_PARENT_PARENT_DIRS = {
    "docparams",
    "deprecated_relative_import",
    "ext",
}
"""Parents of direct parent directories that should be ignored."""


def get_functional_test_files_from_directory(
    input_dir: Path | str,
    max_file_per_directory: int = REASONABLY_DISPLAYABLE_VERTICALLY,
) -> list[FunctionalTestFile]:
    """Get all functional tests in the input_dir."""
    suite = []

    _check_functional_tests_structure(Path(input_dir), max_file_per_directory)

    for dirpath, dirnames, filenames in os.walk(input_dir):
        if dirpath.endswith("__pycache__"):
            continue
        dirnames.sort()
        filenames.sort()
        for filename in filenames:
            if filename != "__init__.py" and filename.endswith(".py"):
                suite.append(FunctionalTestFile(dirpath, filename))
    return suite


def _check_functional_tests_structure(
    directory: Path, max_file_per_directory: int
) -> None:
    """Check if test directories follow correct file/folder structure.

    Ignore underscored directories or files.
    """
    if Path(directory).stem.startswith("_"):
        return

    files: set[Path] = set()
    dirs: set[Path] = set()

    def _get_files_from_dir(
        path: Path, violations: list[tuple[Path, int]]
    ) -> list[Path]:
        """Return directories and files from a directory and handles violations."""
        files_without_leading_underscore = list(
            p
            for p in path.iterdir()
            if not (p.stem.startswith("_") or (p.is_file() and p.suffix != ".py"))
        )
        if len(files_without_leading_underscore) > max_file_per_directory:
            violations.append((path, len(files_without_leading_underscore)))
        return files_without_leading_underscore

    def walk(path: Path) -> Iterator[Path]:
        violations: list[tuple[Path, int]] = []
        violations_msgs: set[str] = set()
        parent_dir_files = _get_files_from_dir(path, violations)
        error_msg = (
            "The following directory contains too many functional tests files:\n"
        )
        for _file_or_dir in parent_dir_files:
            if _file_or_dir.is_dir():
                yield _file_or_dir.resolve()
                try:
                    yield from walk(_file_or_dir)
                except AssertionError as e:
                    violations_msgs.add(str(e).replace(error_msg, ""))
            else:
                yield _file_or_dir.resolve()
        if violations or violations_msgs:
            _msg = error_msg
            for offending_file, number in violations:
                _msg += f"- {offending_file}: {number} when the max is {max_file_per_directory}\n"
            for error_msg in violations_msgs:
                _msg += error_msg
            raise AssertionError(_msg)

    # Collect all sub-directories and files in directory
    for file_or_dir in walk(directory):
        if file_or_dir.is_dir():
            dirs.add(file_or_dir)
        elif file_or_dir.suffix == ".py":
            files.add(file_or_dir)

    directory_does_not_exists: list[tuple[Path, Path]] = []
    misplaced_file: list[Path] = []
    for file in files:
        possible_dir = file.parent / file.stem.split("_")[0]
        if possible_dir.exists():
            directory_does_not_exists.append((file, possible_dir))
        # Exclude some directories as they follow a different structure
        if (
            not len(file.parent.stem) == 1  # First letter sub-directories
            and file.parent.stem not in IGNORED_PARENT_DIRS
            and file.parent.parent.stem not in IGNORED_PARENT_PARENT_DIRS
        ):
            if not file.stem.startswith(file.parent.stem):
                misplaced_file.append(file)

    if directory_does_not_exists or misplaced_file:
        msg = "The following functional tests are disorganized:\n"
        for file, possible_dir in directory_does_not_exists:
            msg += (
                f"- In '{directory}', '{file.relative_to(directory)}' "
                f"should go in '{possible_dir.relative_to(directory)}'\n"
            )
        for file in misplaced_file:
            msg += (
                f"- In '{directory}', {file.relative_to(directory)} should go in a directory"
                f" that starts with the first letters"
                f" of '{file.stem}' (not '{file.parent.stem}')\n"
            )
        raise AssertionError(msg)



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pylint.checkers.typecheck import _similar_names
from pylint.checkers.utils import (
    class_is_abstract,
    in_for_else_branch,
    infer_all,
    is_overload_stub,
    overridden_method,
    safe_infer,
    unimplemented_abstract_methods,
)

if TYPE_CHECKING:
    from functools import _lru_cache_wrapper


def clear_lru_caches() -> None:
    """Clear caches holding references to AST nodes."""
    caches_holding_node_references: list[_lru_cache_wrapper[Any]] = [
        class_is_abstract,
        in_for_else_branch,
        infer_all,
        is_overload_stub,
        overridden_method,
        unimplemented_abstract_methods,
        safe_infer,
        _similar_names,
    ]
    for lru in caches_holding_node_references:
        lru.cache_clear()



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import itertools

import astroid
from astroid import bases, nodes, util

from pylint import checkers
from pylint.checkers import utils
from pylint.interfaces import HIGH, INFERENCE


def _is_constant_zero(node: str | nodes.NodeNG) -> bool:
    # We have to check that node.value is not False because node.value == 0 is True
    # when node.value is False
    return isinstance(node, nodes.Const) and node.value == 0 and node.value is not False


class ImplicitBooleanessChecker(checkers.BaseChecker):
    """Checks for incorrect usage of comparisons or len() inside conditions.

    Incorrect usage of len()
    Pep8 states:
    For sequences, (strings, lists, tuples), use the fact that empty sequences are false.

        Yes: if not seq:
             if seq:

        No: if len(seq):
            if not len(seq):

    Problems detected:
    * if len(sequence):
    * if not len(sequence):
    * elif len(sequence):
    * elif not len(sequence):
    * while len(sequence):
    * while not len(sequence):
    * assert len(sequence):
    * assert not len(sequence):
    * bool(len(sequence))

    Incorrect usage of empty literal sequences; (), [], {},

    For empty sequences, (dicts, lists, tuples), use the fact that empty sequences are false.

        Yes: if variable:
             if not variable

        No: if variable == empty_literal:
            if variable != empty_literal:

    Problems detected:
    * comparison such as variable == empty_literal:
    * comparison such as variable != empty_literal:
    """

    name = "refactoring"
    msgs = {
        "C1802": (
            "Do not use `len(SEQUENCE)` without comparison to determine if a sequence is empty",
            "use-implicit-booleaness-not-len",
            "Empty sequences are considered false in a boolean context. You can either"
            " remove the call to 'len' (``if not x``) or compare the length against a"
            " scalar (``if len(x) > 1``).",
            {"old_names": [("C1801", "len-as-condition")]},
        ),
        "C1803": (
            '"%s" can be simplified to "%s", if it is strictly a sequence, as an empty %s is falsey',
            "use-implicit-booleaness-not-comparison",
            "Empty sequences are considered false in a boolean context. Following this"
            " check blindly in weakly typed code base can create hard to debug issues."
            " If the value can be something else that is falsey but not a sequence (for"
            " example ``None``, an empty string, or ``0``) the code will not be "
            "equivalent.",
        ),
        "C1804": (
            '"%s" can be simplified to "%s", if it is strictly a string, as an empty string is falsey',
            "use-implicit-booleaness-not-comparison-to-string",
            "Empty string are considered false in a boolean context. Following this"
            " check blindly in weakly typed code base can create hard to debug issues."
            " If the value can be something else that is falsey but not a string (for"
            " example ``None``, an empty sequence, or ``0``) the code will not be "
            "equivalent.",
            {
                "default_enabled": False,
                "old_names": [("C1901", "compare-to-empty-string")],
            },
        ),
        "C1805": (
            '"%s" can be simplified to "%s", if it is strictly an int, as 0 is falsey',
            "use-implicit-booleaness-not-comparison-to-zero",
            "0 is considered false in a boolean context. Following this"
            " check blindly in weakly typed code base can create hard to debug issues."
            " If the value can be something else that is falsey but not an int (for"
            " example ``None``, an empty string, or an empty sequence) the code will not be "
            "equivalent.",
            {"default_enabled": False, "old_names": [("C2001", "compare-to-zero")]},
        ),
    }

    options = ()
    _operators = {"!=", "==", "is not", "is"}

    @utils.only_required_for_messages("use-implicit-booleaness-not-len")
    def visit_call(self, node: nodes.Call) -> None:
        # a len(S) call is used inside a test condition
        # could be if, while, assert or if expression statement
        # e.g. `if len(S):`
        if not utils.is_call_of_name(node, "len"):
            return
        # the len() call could also be nested together with other
        # boolean operations, e.g. `if z or len(x):`
        parent = node.parent
        while isinstance(parent, nodes.BoolOp):
            parent = parent.parent
        # we're finally out of any nested boolean operations so check if
        # this len() call is part of a test condition
        if not utils.is_test_condition(node, parent):
            return
        if not node.args:
            return
        len_arg = node.args[0]
        if isinstance(len_arg, (nodes.ListComp, nodes.SetComp, nodes.DictComp)):
            # The node is a comprehension as in len([x for x in ...])
            self.add_message(
                "use-implicit-booleaness-not-len",
                node=node,
                confidence=HIGH,
            )
            return
        try:
            instance = next(len_arg.infer())
        except astroid.InferenceError:
            # Probably undefined-variable, abort check
            return
        mother_classes = self.base_names_of_instance(instance)
        affected_by_pep8 = any(
            t in mother_classes for t in ("str", "tuple", "list", "set")
        )
        if "range" in mother_classes or (
            affected_by_pep8 and not self.instance_has_bool(instance)
        ):
            self.add_message(
                "use-implicit-booleaness-not-len",
                node=node,
                confidence=INFERENCE,
            )

    @staticmethod
    def instance_has_bool(class_def: nodes.ClassDef) -> bool:
        try:
            class_def.getattr("__bool__")
            return True
        except astroid.AttributeInferenceError:
            ...
        return False

    @utils.only_required_for_messages("use-implicit-booleaness-not-len")
    def visit_unaryop(self, node: nodes.UnaryOp) -> None:
        """`not len(S)` must become `not S` regardless if the parent block is a test
        condition or something else (boolean expression) e.g. `if not len(S):`.
        """
        if (
            isinstance(node, nodes.UnaryOp)
            and node.op == "not"
            and utils.is_call_of_name(node.operand, "len")
        ):
            self.add_message(
                "use-implicit-booleaness-not-len", node=node, confidence=HIGH
            )

    @utils.only_required_for_messages(
        "use-implicit-booleaness-not-comparison",
        "use-implicit-booleaness-not-comparison-to-string",
        "use-implicit-booleaness-not-comparison-to-zero",
    )
    def visit_compare(self, node: nodes.Compare) -> None:
        if self.linter.is_message_enabled("use-implicit-booleaness-not-comparison"):
            self._check_use_implicit_booleaness_not_comparison(node)
        if self.linter.is_message_enabled(
            "use-implicit-booleaness-not-comparison-to-zero"
        ) or self.linter.is_message_enabled(
            "use-implicit-booleaness-not-comparison-to-str"
        ):
            self._check_compare_to_str_or_zero(node)

    def _check_compare_to_str_or_zero(self, node: nodes.Compare) -> None:
        # Skip check for chained comparisons
        if len(node.ops) != 1:
            return

        negation_redundant_ops = {"!=", "is not"}
        # note: nodes.Compare has the left most operand in node.left
        # while the rest are a list of tuples in node.ops
        # the format of the tuple is ('compare operator sign', node)
        # here we squash everything into `ops` to make it easier for processing later
        ops: list[tuple[str, nodes.NodeNG]] = [("", node.left), *node.ops]
        iter_ops = iter(ops)
        all_ops = list(itertools.chain(*iter_ops))
        _, left_operand, operator, right_operand = all_ops

        if operator not in self._operators:
            return
        if self.linter.is_message_enabled(
            "use-implicit-booleaness-not-comparison-to-zero"
        ):
            operand = None
            # 0 ?? X
            if _is_constant_zero(left_operand):
                operand = right_operand
            # X ?? 0
            elif _is_constant_zero(right_operand):
                operand = left_operand
            if operand is not None:
                original = (
                    f"{left_operand.as_string()} {operator} {right_operand.as_string()}"
                )
                suggestion = self._get_suggestion(
                    node, operand.as_string(), operator, negation_redundant_ops
                )
                self.add_message(
                    "use-implicit-booleaness-not-comparison-to-zero",
                    args=(original, suggestion),
                    node=node,
                    confidence=HIGH,
                )
        if self.linter.is_message_enabled(
            "use-implicit-booleaness-not-comparison-to-str"
        ):
            node_name = None
            # x ?? ""
            if utils.is_empty_str_literal(left_operand):
                node_name = right_operand.as_string()
            # '' ?? X
            elif utils.is_empty_str_literal(right_operand):
                node_name = left_operand.as_string()
            if node_name is not None:
                suggestion = self._get_suggestion(
                    node, node_name, operator, negation_redundant_ops
                )
                self.add_message(
                    "use-implicit-booleaness-not-comparison-to-string",
                    args=(node.as_string(), suggestion),
                    node=node,
                    confidence=HIGH,
                )

    def _check_use_implicit_booleaness_not_comparison(
        self, node: nodes.Compare
    ) -> None:
        """Check for left side and right side of the node for empty literals."""
        # Skip check for chained comparisons
        if len(node.ops) != 1:
            return

        # Check both left-hand side and right-hand side for literals
        operator, comparator = node.ops[0]
        is_left_empty_literal = utils.is_base_container(
            node.left
        ) or utils.is_empty_dict_literal(node.left)
        is_right_empty_literal = utils.is_base_container(
            comparator
        ) or utils.is_empty_dict_literal(comparator)

        # If both sides are literals/non-literals, it should be different error.
        if not (is_left_empty_literal ^ is_right_empty_literal):
            return

        # Set target_node to opposite side of literal
        target_node = node.left if is_right_empty_literal else comparator
        literal_node = comparator if is_right_empty_literal else node.left
        # Infer node to check
        target_instance = utils.safe_infer(target_node)
        if target_instance is None:
            return
        mother_classes = self.base_names_of_instance(target_instance)
        is_base_comprehension_type = any(
            t in mother_classes for t in ("tuple", "list", "dict", "set")
        )

        # Only time we bypass check is when target_node is not inherited by
        # collection literals and have its own __bool__ implementation.
        if not is_base_comprehension_type and self.instance_has_bool(target_instance):
            return

        # No need to check for operator when visiting compare node
        if operator in {"==", "!=", ">=", ">", "<=", "<"}:
            self.add_message(
                "use-implicit-booleaness-not-comparison",
                args=self._implicit_booleaness_message_args(
                    node, literal_node, operator, target_node
                ),
                node=node,
                confidence=HIGH,
            )

    def _get_node_description(self, node: nodes.NodeNG) -> str:
        return {
            nodes.List: "list",
            nodes.Tuple: "tuple",
            nodes.Dict: "dict",
            nodes.Const: "str",
        }.get(type(node), "iterable")

    def _implicit_booleaness_message_args(
        self,
        node: nodes.Compare,
        literal_node: nodes.NodeNG,
        operator: str,
        target_node: nodes.NodeNG,
    ) -> tuple[str, str, str]:
        """Helper to get the right message for "use-implicit-booleaness-not-comparison"."""
        description = self._get_node_description(literal_node)
        collection_literal = {
            "list": "[]",
            "tuple": "()",
            "dict": "{}",
        }.get(description, "iterable")
        instance_name = "x"
        match target_node:
            case nodes.Call():
                instance_name = f"{target_node.func.as_string()}(...)"
            case nodes.Attribute() | nodes.Name():
                instance_name = target_node.as_string()
        original_comparison = f"{instance_name} {operator} {collection_literal}"
        suggestion = self._get_suggestion(node, instance_name, operator, {"!="})
        return original_comparison, suggestion, description

    def _get_suggestion(
        self,
        node: nodes.Compare,
        name: str,
        operator: str,
        negation_redundant_ops: set[str],
    ) -> str:
        if operator in negation_redundant_ops:
            return f"{name}" if self._in_boolean_context(node) else f"bool({name})"
        return f"not {name}"

    def _in_boolean_context(self, node: nodes.Compare) -> bool:
        """Returns True if the comparison is used in a boolean context; False otherwise.

        A comparison is considered to be in a boolean context when it appears in constructs
        that evaluate its truthiness directly, such as:
        - control flow statements (`if`, `while`, `assert`)
        - ternary expressions (`x if condition else y`)
        - logical negation (`not`)
        - comprehension filters (e.g. `[x for x in items if cond]`)
        - generator expressions passed to `all()` or `any()`
        - lambdas expressions passed to `filter()`
        - `bool()` cast
        - boolean operations `and`, `or` nested within any of the above

        In contrast, a comparison is not in a boolean context when its result is used as a value,
        such as when it is assigned to a variable, returned from a function, passed as an argument,
        or used in an expression that does not depend on its truthiness.
        """
        current, parent = node, node.parent
        while parent:
            match parent:
                case nodes.If() | nodes.While() | nodes.Assert() if (
                    current is parent.test
                ):
                    return True
                case nodes.IfExp() if current is parent.test:
                    return True

                case nodes.UnaryOp(op="not") if current is parent.operand:
                    return True

                case nodes.Comprehension() if current in parent.ifs:
                    return True

                case nodes.GeneratorExp() if (
                    current is parent.elt
                    and (
                        utils.is_call_of_name(parent.parent, "all")
                        or utils.is_call_of_name(parent.parent, "any")
                    )
                    and parent in parent.parent.args
                ):
                    return True

                case nodes.Lambda() if (
                    current is parent.body
                    and utils.is_call_of_name(parent.parent, "filter")
                    and parent in parent.parent.args
                ):
                    return True

                case _ if (
                    utils.is_call_of_name(parent, "bool") and current in parent.args
                ):
                    return True

                case nodes.BoolOp() if current in parent.values:
                    current = parent
                    parent = current.parent
                    continue

                case _:
                    break

        return False

    @staticmethod
    def base_names_of_instance(
        node: util.UninferableBase | bases.Instance,
    ) -> list[str]:
        """Return all names inherited by a class instance or those returned by a
        function.

        The inherited names include 'object'.
        """
        if isinstance(node, bases.Instance):
            return [node.name] + [x.name for x in node.ancestors()]
        return []



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Create UML diagrams for classes and modules in <packages>."""

from __future__ import annotations

import sys
from collections.abc import Sequence

from pylint import constants
from pylint.config.arguments_manager import _ArgumentsManager
from pylint.config.arguments_provider import _ArgumentsProvider
from pylint.lint import discover_package_path
from pylint.lint.utils import augmented_sys_path
from pylint.pyreverse import writer
from pylint.pyreverse.diadefslib import DiadefsHandler
from pylint.pyreverse.inspector import Linker, project_from_files
from pylint.pyreverse.utils import (
    check_graphviz_availability,
    check_if_graphviz_supports_format,
    insert_default_options,
)
from pylint.typing import Options

DIRECTLY_SUPPORTED_FORMATS = (
    "dot",
    "puml",
    "plantuml",
    "mmd",
    "html",
)

DEFAULT_COLOR_PALETTE = (
    # colorblind scheme taken from https://personal.sron.nl/~pault/
    "#77AADD",  # light blue
    "#99DDFF",  # light cyan
    "#44BB99",  # mint
    "#BBCC33",  # pear
    "#AAAA00",  # olive
    "#EEDD88",  # light yellow
    "#EE8866",  # orange
    "#FFAABB",  # pink
    "#DDDDDD",  # pale grey
)


OPTIONS_GROUPS = {
    "FILTERING": "Filtering and Scope",
    "DISPLAY": "Display Options",
    "OUTPUT": "Output Control",
    "PROJECT": "Project Configuration",
}


OPTIONS: Options = (
    # Filtering and Scope options
    (
        "filter-mode",
        {
            "short": "f",
            "default": "PUB_ONLY",
            "dest": "mode",
            "type": "string",
            "action": "store",
            "metavar": "<mode>",
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": """Filter attributes and functions according to <mode>. Correct modes are:
'PUB_ONLY' filter all non public attributes [DEFAULT], equivalent to PRIVATE+SPECIAL
'ALL' no filter
'SPECIAL' filter Python special functions except constructor
'OTHER' filter protected and private attributes""",
        },
    ),
    (
        "class",
        {
            "short": "c",
            "action": "extend",
            "metavar": "<class>",
            "type": "csv",
            "dest": "classes",
            "default": None,
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": "Create a class diagram with all classes related to <class>;\
 this uses by default the options -ASmy",
        },
    ),
    (
        "show-ancestors",
        {
            "short": "a",
            "action": "store",
            "metavar": "<ancestor>",
            "type": "int",
            "default": None,
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": "Show <ancestor> generations of ancestor classes not in <projects>.",
        },
    ),
    (
        "all-ancestors",
        {
            "short": "A",
            "default": None,
            "action": "store_true",
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": "Show all ancestors of all classes in <projects>.",
        },
    ),
    (
        "show-associated",
        {
            "short": "s",
            "action": "store",
            "metavar": "<association_level>",
            "type": "int",
            "default": None,
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": "Show <association_level> levels of associated classes not in <projects>.",
        },
    ),
    (
        "all-associated",
        {
            "short": "S",
            "default": None,
            "action": "store_true",
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": "Show all classes associated with the target classes, including indirect associations.",
        },
    ),
    (
        "show-builtin",
        {
            "short": "b",
            "action": "store_true",
            "default": False,
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": "Include builtin objects in representation of classes.",
        },
    ),
    (
        "show-stdlib",
        {
            "short": "L",
            "action": "store_true",
            "default": False,
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": "Include standard library objects in representation of classes.",
        },
    ),
    (
        "max-depth",
        {
            "dest": "max_depth",
            "action": "store",
            "default": None,
            "metavar": "<depth>",
            "type": "int",
            "group": OPTIONS_GROUPS["FILTERING"],
            "help": (
                "Maximum depth of packages/modules to include in the diagram, relative to the "
                "deepest specified package. A depth of 0 shows only the specified packages/modules, "
                "while 1 includes their immediate children, etc. When specifying nested packages,  "
                "depth is calculated from the deepest package level. If not specified, all "
                "packages/modules in the hierarchy are shown."
            ),
        },
    ),
    # Display Options
    (
        "module-names",
        {
            "short": "m",
            "default": None,
            "type": "yn",
            "metavar": "<y or n>",
            "group": OPTIONS_GROUPS["DISPLAY"],
            "help": "Include module name in the representation of classes.",
        },
    ),
    (
        "only-classnames",
        {
            "short": "k",
            "action": "store_true",
            "default": False,
            "group": OPTIONS_GROUPS["DISPLAY"],
            "help": "Don't show attributes and methods in the class boxes; this disables -f values.",
        },
    ),
    (
        "no-standalone",
        {
            "action": "store_true",
            "default": False,
            "group": OPTIONS_GROUPS["DISPLAY"],
            "help": "Only show nodes with connections.",
        },
    ),
    (
        "colorized",
        {
            "dest": "colorized",
            "action": "store_true",
            "default": False,
            "group": OPTIONS_GROUPS["DISPLAY"],
            "help": "Use colored output. Classes/modules of the same package get the same color.",
        },
    ),
    (
        "max-color-depth",
        {
            "dest": "max_color_depth",
            "action": "store",
            "default": 2,
            "metavar": "<depth>",
            "type": "int",
            "group": OPTIONS_GROUPS["DISPLAY"],
            "help": "Use separate colors up to package depth of <depth>. Higher depths will reuse colors.",
        },
    ),
    (
        "color-palette",
        {
            "dest": "color_palette",
            "action": "store",
            "default": DEFAULT_COLOR_PALETTE,
            "metavar": "<color1,color2,...>",
            "type": "csv",
            "group": OPTIONS_GROUPS["DISPLAY"],
            "help": "Comma separated list of colors to use for the package depth coloring.",
        },
    ),
    # Output Control options
    (
        "output",
        {
            "short": "o",
            "dest": "output_format",
            "action": "store",
            "default": "dot",
            "metavar": "<format>",
            "type": "string",
            "group": OPTIONS_GROUPS["OUTPUT"],
            "help": (
                "Create a *.<format> output file if format is available. Available "
                f"formats are: {', '.join('.' + fmt for fmt in DIRECTLY_SUPPORTED_FORMATS)}. Any other "
                "format will be tried to be created by using the 'dot' command line "
                "tool, which requires a graphviz installation. In this case, these additional "
                "formats are available (see `Graphviz output formats <https://graphviz.org/docs/outputs/>`_)."
            ),
        },
    ),
    (
        "output-directory",
        {
            "default": "",
            "type": "path",
            "short": "d",
            "action": "store",
            "metavar": "<output_directory>",
            "group": OPTIONS_GROUPS["OUTPUT"],
            "help": "Set the output directory path.",
        },
    ),
    # Project Configuration options
    (
        "ignore",
        {
            "type": "csv",
            "metavar": "<file[,file...]>",
            "dest": "ignore_list",
            "default": constants.DEFAULT_IGNORE_LIST,
            "group": OPTIONS_GROUPS["PROJECT"],
            "help": "Files or directories to be skipped. They should be base names, not paths.",
        },
    ),
    (
        "project",
        {
            "default": "",
            "type": "string",
            "short": "p",
            "metavar": "<project name>",
            "group": OPTIONS_GROUPS["PROJECT"],
            "help": "Set the project name. This will later be appended to the output file names.",
        },
    ),
    (
        "source-roots",
        {
            "type": "glob_paths_csv",
            "metavar": "<path>[,<path>...]",
            "default": (),
            "group": OPTIONS_GROUPS["PROJECT"],
            "help": "Add paths to the list of the source roots. Supports globbing patterns. The "
            "source root is an absolute path or a path relative to the current working directory "
            "used to determine a package namespace for modules located under the source root.",
        },
    ),
    (
        "verbose",
        {
            "action": "store_true",
            "default": False,
            "group": OPTIONS_GROUPS["PROJECT"],
            "help": "Makes pyreverse more verbose/talkative. Mostly useful for debugging.",
        },
    ),
)


# Base class providing common behaviour for pyreverse commands
class Run(_ArgumentsManager, _ArgumentsProvider):
    options = OPTIONS
    name = "pyreverse"

    def __init__(self, args: Sequence[str]) -> None:
        # Immediately exit if user asks for version
        if "--version" in args:
            print("pyreverse is included in pylint:")
            print(constants.full_version)
            sys.exit(0)

        _ArgumentsManager.__init__(self, prog="pyreverse", description=__doc__)
        _ArgumentsProvider.__init__(self, self)

        # Parse options
        insert_default_options()
        self.args = self._parse_command_line_configuration(args)

        if self.config.output_format not in DIRECTLY_SUPPORTED_FORMATS:
            check_graphviz_availability()
            print(
                f"Format {self.config.output_format} is not supported natively."
                " Pyreverse will try to generate it using Graphviz..."
            )
            check_if_graphviz_supports_format(self.config.output_format)

    def run(self) -> int:
        """Checking arguments and run project."""
        if not self.args:
            print(self.help())
            return 1
        extra_packages_paths = list(
            {discover_package_path(arg, self.config.source_roots) for arg in self.args}
        )
        with augmented_sys_path(extra_packages_paths):
            project = project_from_files(
                self.args,
                project_name=self.config.project,
                black_list=self.config.ignore_list,
                verbose=self.config.verbose,
            )
            linker = Linker(project, tag=True)
            handler = DiadefsHandler(self.config, self.args)
            diadefs = handler.get_diadefs(project, linker)
        writer.DiagramWriter(self.config).write(diadefs)
        return 0


if __name__ == "__main__":
    arguments = sys.argv[1:]
    Run(arguments).run()



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

from itertools import zip_longest
from typing import TYPE_CHECKING

from astroid import nodes

from pylint.checkers import BaseChecker
from pylint.interfaces import HIGH

if TYPE_CHECKING:
    from pylint.lint import PyLinter


class LambdaExpressionChecker(BaseChecker):
    """Check for unnecessary usage of lambda expressions."""

    name = "lambda-expressions"
    msgs = {
        "C3001": (
            "Lambda expression assigned to a variable. "
            'Define a function using the "def" keyword instead.',
            "unnecessary-lambda-assignment",
            "Used when a lambda expression is assigned to variable "
            'rather than defining a standard function with the "def" keyword.',
        ),
        "C3002": (
            "Lambda expression called directly. Execute the expression inline instead.",
            "unnecessary-direct-lambda-call",
            "Used when a lambda expression is directly called "
            "rather than executing its contents inline.",
        ),
    }
    options = ()

    def visit_assign(self, node: nodes.Assign) -> None:
        """Check if lambda expression is assigned to a variable."""
        match node:
            case nodes.Assign(
                targets=[nodes.AssignName(), *_], value=nodes.Lambda() as value
            ):
                self.add_message(
                    "unnecessary-lambda-assignment",
                    node=value,
                    confidence=HIGH,
                )
            case nodes.Assign(
                targets=[nodes.Tuple() as target, *_],
                value=nodes.Tuple() | nodes.List() as value,
            ):
                # Iterate over tuple unpacking assignment elements and
                # see if any lambdas are assigned to a variable.
                # N.B. We may encounter W0632 (unbalanced-tuple-unpacking)
                # and still need to flag the lambdas that are being assigned.
                for lhs_elem, rhs_elem in zip_longest(target.elts, value.elts):
                    if lhs_elem is None or rhs_elem is None:
                        # unbalanced tuple unpacking. stop checking.
                        break
                    if isinstance(lhs_elem, nodes.AssignName) and isinstance(
                        rhs_elem, nodes.Lambda
                    ):
                        self.add_message(
                            "unnecessary-lambda-assignment",
                            node=rhs_elem,
                            confidence=HIGH,
                        )

    def visit_namedexpr(self, node: nodes.NamedExpr) -> None:
        match node:
            case nodes.NamedExpr(
                target=nodes.AssignName(), value=nodes.Lambda() as value
            ):
                self.add_message(
                    "unnecessary-lambda-assignment",
                    node=value,
                    confidence=HIGH,
                )

    def visit_call(self, node: nodes.Call) -> None:
        """Check if lambda expression is called directly."""
        if isinstance(node.func, nodes.Lambda):
            self.add_message(
                "unnecessary-direct-lambda-call",
                node=node,
                confidence=HIGH,
            )


def register(linter: PyLinter) -> None:
    linter.register_checker(LambdaExpressionChecker(linter))
